# -*- coding: utf-8 -*-

"""
`Beam` class and associated utilities
"""

import time
import uuid

from pewpew import _context as ctx

__all__ = [
    "Beam",
]


class Beam:
    """A context manager for performance timings.

    A context manager that can be used to time a block's performance.
    The `Span` works by storing the time relative to program start
    (see: `time.perf_counter`) as a starting pointer on `__enter__`,
    and then the new relative time on `__exit__`. Iteration timings are
    dynamically calculated with the `timings` property.

    NOTE: a span should not be accessed in concurrent fashion. A
    new span should be created for each block that is executed concurrently.

    Examples
    --------
    >>> import pewpew as pw
    >>> span = pw.Beam("big_loop")
    >>>
    >>> with span:
    ...     for i in range(1000000):
    ...         pass
    ...
    >>> span.timings
    ... (0.0386255190000071,)
    >>> with span:
    ...     for i in range(1000000):
    ...         pass
    ...
    >>> span.timings
    ... (0.0386255190000071, 0.039705107999992606)

    Parameters
    ----------
    name : str
        The name of the span. A `Beam`'s name does not need to be unique,
        as each `Beam` contains a unique identifier. The name of a `Beam`
        should describe the logical block it spans. For instance, a beam
        tracing an expensive I/O operation should describe the operation
        concisely. However, a `Beam` created in a loop over an operation
        does not need to be named with a trailing `i` value, for instance.
    """

    def __init__(self, name):
        self.name = name
        self.clear()
        self._id = str(uuid.uuid4())

        # if this was created inside the context of `@pewpew.trace` we want
        # to track the Beam within the scope of that context
        ctx.ContextStore.track_beam(self)

    def clear(self):
        """Completely clear the history of the span, resetting to its zero state"""
        self._reset_counter()
        self._start_times = []
        self._end_times = []

    def _reset_counter(self):
        """A span can be re-used, and resets on `__enter__`"""
        self._start = self._stop = None

    def __enter__(self):
        """Enter the context manager"""
        self._reset_counter()
        self._start = time.perf_counter()
        self._start_times.append(self._start)
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        """Exit the context manager"""
        stop = time.perf_counter()
        self._end_times.append(stop)
        self._reset_counter()

    def _first_start_time(self):
        """Get the min value of the start array"""
        return min(self._start_times)

    def _last_end_time(self):
        """Get the max value of the end array"""
        return max(self._end_times)

    def _zipped(self):
        """Zip start/end times + concurrency check"""
        starts = self._start_times
        stops = self._end_times

        # detect whether someone tries to access this from within
        # the context manager block
        if len(starts) != len(stops):
            raise RuntimeError(
                f"Concurrent access to {self.__class__.__name__} from within "
                "context manager block"
            )
        return zip(starts, stops)

    @property
    def timings(self):
        """Get a span's timing history"""
        starts_stops = self._zipped()
        return tuple(map(lambda ab: ab[1] - ab[0], starts_stops))



# -*- coding: utf-8 -*-

"""
The global context. Collects all `Beam` traces while active
"""

import collections
import copy
import os
import threading

from pewpew._logging import get_logger
from pewpew.utils import iterables as iter_utils

__all__ = [
    "clear_trace_history",
]

logger = get_logger(__name__)

# Used to serialize access to shared context data structures in `pewpew`.
# TODO: @TayTay -- consider whether should be a `Lock` instead
_lock = threading.RLock()
_int_safe = lambda _x: int(_x) if _x is not None else None
_get_hist_len = lambda: _int_safe(os.environ.get("PEWPEW_TRACE_HISTORY_SIZE", None))


class _ContextStore:
    """A manager class that internally stores parent/child
    relationships between contexts

    This class allows multiple nested `pewpew.trace` calls to stack
    and for the `Beam`s created within those contexts to be tracked.
    The context store is implemented as a linked list, maintaining a
    pointer to the root, as well as the head (the active context).
    When all contexts have been exited, and there is no active context,
    the root is placed into an internal list to trace trace history
    chronologically.

    This is a class that relies heavily on internal state, and generally
    should not be accessed outside of this module. For public access,
    decorate your function with `@pewpew.trace`. Directly manipulating
    this instance can have unintended and undefined results.

    Note that the `_ContextStore` is managed as a singleton instance,
    which will not cross thread boundaries [1]. If your `pewpew.trace`-
    decorated function spawns `Beam`s within parallel processes internally,
    these will currently not be natively tracked. In these cases, it is
    best to create a `Beam` outside of the context of any multiprocessing.

    Parameters
    ----------
    history_size : int or None, optional
        The size of the trace history deque. This is controlled by the
        `PEWPEW_TRACE_HISTORY_SIZE` environment variable. If not defined,
        the deque will be allowed to grow to an arbitrary length. If set,
        once reached and exceeded, a corresponding number of elements from
        the front of the deque (old traces) will be dropped.

    Notes
    -----
    * This is a private class to organize internal state. Directly
      manipulating this instance can have unintended and undefined
      results.

    References
    ----------
    .. [1] : https://stackoverflow.com/a/45077772/3015734

    Attributes
    ----------
    _root : TraceContext
    _head : TraceContext
    _trace_history : list
    """

    def __init__(self, history_size=None):
        logger.debug(
            f"Creating new context store for thread: {threading.currentThread().ident}"
        )
        self._reset(history_size)
        self._maxlen = history_size

    def _reset(self, history_size):
        """Private method primarily used during testing"""
        with _lock:
            self._root = None  # handle to parent trace history node
            self._head = None  # handle to the tip of the nodes
            self._trace_history = collections.deque(maxlen=history_size)

    def push(self, trace_context):
        """(Potentially create) and mark a child context active

        This is called when a `TraceContext` is entered. The `@pewpew.trace`
        decorator internally creates and uses a `TraceContext`, but they
        can also be manually created.
        """
        with _lock:
            logger.debug(
                f"Pushing new `TraceContext` for thread: {threading.currentThread().ident}"
            )
            # no active context, outer decorator
            if self._root is None:
                self._root = self._head = trace_context
            else:
                # add child to existing parent, update head
                self._head.link_child(trace_context)
                self._head = trace_context

    def pop(self):
        """Pop the head of the store

        This is called when a `TraceContext` is exited. The `@pewpew.trace`
        decorator internally creates and uses a `TraceContext`, but they
        can also be manually created.
        """
        with _lock:
            logger.debug(
                f"Popping `TraceContext` for thread: {threading.currentThread().ident}"
            )
            # Only happens if external user is messing with this
            if self._head is None:
                return

            head = self._head

            # NOTE: we do NOT clear the handle to the child so the beam
            # history is retained after the decorator is exited. If we
            # cleared the `_child` pointer, we'd not be able to recover
            # the traces. We only push back the pointer to `_head`
            self._head = head._parent

            # Base case: if head is now None, `head` was the root. We
            # want to retain the history of the code path / traces, so
            # we push the parent into the history.
            if self._head is None:
                self._root = None
                self._trace_history.append(head)

    def track_beam(self, beam):
        """If an active context is present, track a new beam"""
        with _lock:
            if self._head is not None:
                self._head._track_beam(beam)

    def get_trace(self, name=None, idx=None):
        """Find a trace by name or index

        If no args are supplied, returns the most recent trace. If
        a `name` is provided that matches multiple traces, the most
        recent will be returned.
        """
        if name is None and idx is None:
            idx = -1

        beams = []
        with _lock:
            if self._trace_history:
                if idx is not None:
                    trace = self._trace_history[idx]
                else:
                    trace = iter_utils.find_first(
                        self._trace_history[::-1],  # reverse order, recent first
                        lambda t: t.name == name,
                    )

                # else flatten the linked list into beams
                while trace is not None:
                    beams.extend(trace.beams)
                    trace = trace._child

                return beams

            return beams


def clear_trace_history():
    """Clear all trace history from the `ContextStore`

    A public method that alters the global state of `pewpew`'s `ContextStore`.
    This function will blow away the trace history, which can be useful in
    preventing any very large trace history from growing the program's memory
    footprint.
    """
    with _lock:
        # drop existing state, create new
        global ContextStore
        ContextStore = _ContextStore(_get_hist_len())


# Singleton instance
ContextStore = _ContextStore(_get_hist_len())


class TraceContext:
    """A context manager to capture and track `Beam`s created while active

    The `TraceContext` is a context manager class entered when a
    `pewpew.trace` decorated function is invoked. Upon context manager entry,
    the instance is set as the "active" context on the context store. If it
    is the first trace context entered, it is set as the root of what ultimately
    can be thought of as a doubly linked list of parent/child `TraceContext`
    nodes.

    This is a class that relies heavily on internal state, and generally
    should NOT be accessed outside of this module. For public access,
    decorate your function with `@pewpew.trace`. Directly manipulating
    this instance can have unintended and undefined results.

    Parameters
    ----------
    name : str
        A named scope for the trace. Could be the module, or a logical
        block. This name will be prepended to any captured beams when
        drawn.
    """

    def __init__(self, name):
        self._name = name
        self._parent = None
        self._child = None

        # all tracked beams while this context is the active one
        self._beams = []

    def __enter__(self):
        """Enter the trace context

        This context will have a handle appended to the global context
        store, and any created `Beam` instances will be tied to the
        most immediate context handle.
        """
        ContextStore.push(self)
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        """Exit the trace context

        Removes this `TraceContext` from being the "active" context in
        the `ContextStore`.
        """
        ContextStore.pop()

    def link_child(self, child):
        """Assign the child, double linking this as the parent"""
        self._child = child
        self._child._parent = self

    def _track_beam(self, beam):
        """Track a beam assigned in this context"""
        logger.debug(f"Tracking `Beam` for thread: {threading.currentThread().ident}")
        self._beams.append(beam)

    @property
    def beams(self):
        """Return copies of the context's beams pre-pended with the context name"""

        # TODO: consider greater fine grain control over this
        # def _copy_beam(beam):
        #     beam = copy.deepcopy(beam)
        #     beam.name = f"{self.name}.{beam.name}"
        #     return beam
        # return [_copy_beam(b) for b in self._beams]

        return copy.deepcopy(self._beams)

    @property
    def name(self):
        """Return the name of the trace"""
        return self._name



# -*- coding: utf-8 -*-

"""
Useful decorators and utils
"""

import inspect

from pewpew import _context as ctx

__all__ = [
    "make_decorator",
]


class Decorator:
    """Decorated wrapper to carry over metadata from the decorated fn"""

    def __init__(self, func, name):
        self._fn = func
        self._name = name

        if hasattr(func, "__name__"):
            self.__name__ = func.__name__
        if hasattr(func, "__qualname__"):
            self.__qualname__ = func.__qualname__

    def __call__(self, *args, **kwargs):
        """This is the primary entrypoint to the decorated function"""
        # TODO: use a UUID4 rather than name here?
        with ctx.TraceContext(self.name):
            return self._fn(*args, **kwargs)

    @property
    def name(self):
        return self._name


def make_decorator(func, name=None):
    """Public wrapper to create a `Decorator` for decorated functions"""
    if name is None:
        name = inspect.currentframe().f_back.f_code.co_name

    # TODO: get cute with any other attr automation
    return Decorator(func, name=name)



# -*- coding: utf-8 -*-

import pytest
from pewpew.utils import validation as val_utils


@pytest.mark.parametrize(
    "x,default,exp",
    [
        pytest.param(None, None, None),
        pytest.param(None, "dflt", "dflt"),
        pytest.param("non-none", 1, "non-none"),
    ],
)
def test_coalesce(x, default, exp):
    got = val_utils.coalesce(x, default)
    assert exp == got, got



# -*- coding: utf-8 -*-

import os
import pytest
import tempfile

import pewpew
from pewpew import _context as ctx
from pewpew.utils import patching


@pewpew.trace
def no_arg_decorator_add(a, b):
    with pewpew.Beam("add1"):
        return a + b


@pewpew.trace(context_name="TEST_IT_UP_FAM")
def arg_decorator_add(a, b):
    with pewpew.Beam("add1"):
        return a + b


@pytest.mark.parametrize("fn", [no_arg_decorator_add, arg_decorator_add])
def test_diff_decorator_types(fn):
    ctx.clear_trace_history()
    got = fn(1, 2)
    assert got == 3  # just to show nothing is manipulated in the return
    assert ctx.ContextStore._root is ctx.ContextStore._head is None
    assert len(ctx.ContextStore._trace_history) == 1


@pytest.mark.parametrize("fn", [no_arg_decorator_add, arg_decorator_add])
def test_limit_trace_history_len(fn):
    with patching.environ("PEWPEW_TRACE_HISTORY_SIZE", 1):
        ctx.clear_trace_history()
        assert ctx.ContextStore._maxlen == 1

        _ = fn(1, 2)
        assert len(ctx.ContextStore._trace_history) == 1

        # a new call will displace the earlier history here
        _ = fn(1, 2)
        assert len(ctx.ContextStore._trace_history) == 1

    assert "PEWPEW_TRACE_HISTORY_SIZE" not in os.environ
    ctx.clear_trace_history()


@pytest.mark.parametrize("fn", [no_arg_decorator_add, arg_decorator_add])
def test_can_save_fig(fn):
    ctx.clear_trace_history()
    _ = fn(1, 2)

    with tempfile.TemporaryDirectory() as tmp:
        pewpew.draw_trace(save_to=f"{tmp}/fig.png", annotate=True)
        assert os.path.exists(f"{tmp}/fig.png")



# -*- coding: utf-8 -*-

import pewpew
from pewpew import _context as ctx

import unittest


class TestContextStoreDirect(unittest.TestCase):
    """Direct tests of the `ContextStore` and `TraceContext`"""

    @staticmethod
    def assert_context_nodes_empty():
        assert ctx.ContextStore._root is None
        assert ctx.ContextStore._head is None

    def assert_context_empty(self):
        self.assert_context_nodes_empty()
        assert not ctx.ContextStore._trace_history

    def test_beam_no_trace(self):
        ctx.clear_trace_history()
        self.assert_context_empty()

        with pewpew.Beam("test_beam"):
            # `Beam` will NOT be tracked, because not in the context of a TraceContext
            self.assert_context_empty()
        self.assert_context_empty()

    def test_single_trace_context(self):
        ctx.clear_trace_history()
        self.assert_context_empty()

        with ctx.TraceContext(name="test_context") as trace:
            assert ctx.ContextStore._root is trace
            assert ctx.ContextStore._head is trace

            with pewpew.Beam("test_beam"):
                # `Beam` constructor should track the beam in the context
                assert len(ctx.ContextStore._head._beams) == 1

        # when exiting trace context, the trace history should contain the trace now
        assert len(ctx.ContextStore._trace_history) == 1
        self.assert_context_nodes_empty()

    def test_multi_trace_context(self):
        ctx.clear_trace_history()
        self.assert_context_empty()

        with ctx.TraceContext(name="outer_context") as trace1:
            assert ctx.ContextStore._root is trace1
            assert ctx.ContextStore._head is trace1

            with pewpew.Beam("outer_beam"):
                # `Beam` constructor should track the beam in the context
                assert len(ctx.ContextStore._head._beams) == 1

                # inner trace context will Track all new beams on the INNER trace
                with ctx.TraceContext(name="inner_context") as trace2:
                    assert ctx.ContextStore._root is trace1  # root will not change
                    assert ctx.ContextStore._head is trace2

                    with pewpew.Beam("inner_beam1"):
                        assert len(ctx.ContextStore._head._beams) == 1

                    with pewpew.Beam("inner_beam2"):
                        assert len(ctx.ContextStore._head._beams) == 2

        # when exiting trace context, the trace history should contain the trace now
        assert len(ctx.ContextStore._trace_history) == 1
        self.assert_context_nodes_empty()

        # flattening all beams in last trace history should be 3
        beams = ctx.ContextStore.get_trace()
        assert len(beams) == 3



# -*- coding: utf-8 -*-

import contextlib
import os

__all__ = [
    "environ",
]


@contextlib.contextmanager
def environ(k, v):
    """Patch an environment variable for the context of this scope"""
    original = os.environ.get(k, None)
    try:
        os.environ[k] = str(v)
        yield
    finally:
        if original is not None:
            os.environ[k] = original
        elif k in os.environ:
            del os.environ[k]



# -*- coding: utf-8 -*-

"""
The `trace` decorator
"""

from pewpew import _decorator as dec_utils
from pewpew.utils import validation as val_utils

__all__ = [
    "trace",
]


def trace(func=None, context_name="function"):
    """Decorates an outer function, auto-tracing all contained `Beams`

    The primary public `pewpew` utility for tracing, this decorator
    can be applied to any function to automatically collect all created/
    entered `Beam` trace objects during a function's execution.

    Parameters
    ----------
    func : callable
        A python function that contains `pewpew.Beam` traces in its
        stack. For the scope of this function, any `Beam` objects
        created or entered will be auto-detected and traced

    context_name : str, optional
        The name of the trace context. This could be a logical grouping,
        function or module name.

    Examples
    --------
    Ex. 1, an expensive serial function:
    >>> import pewpew
    >>> import matplotlib.pyplot as plt
    >>> import random
    >>> import time
    >>>
    >>> @pewpew.trace
    ... def expensive_serial_function():
    ...     # It's a good idea to declare Beams for anything
    ...     # that's re-used rather than use anonymous ones
    ...     exp_span = pewpew.Beam(name="expensive")
    ...     cheap_span = pewpew.Beam(name="cheap")
    ...
    ...     for _ in range(3):
    ...         with exp_span:
    ...             expensive_thing_that_could_be_parallel()
    ...
    ...     for _ in range(5):
    ...         with cheap_span:
    ...             cheap_serial_thing()
    >>>
    >>> def expensive_thing_that_could_be_parallel():
    ...     time.sleep(3)
    >>>
    >>> def cheap_serial_thing():
    ...     time.sleep(random.uniform(0., 0.25))
    >>>
    >>> expensive_serial_function()
    >>> pewpew.draw_trace(annotate=True)
    >>> plt.show()

    Ex. 2, the same function rethought with multiprocessing:
    >>> import joblib
    >>> import pewpew
    >>> import matplotlib.pyplot as plt
    >>> import random
    >>> import time
    >>>
    >>> @pewpew.trace
    ... def rethought_function(n_jobs=4):
    ...     exp_span = pewpew.Beam(name="expensive_outer")
    ...     with exp_span:
    ...         joblib.Parallel(n_jobs=n_jobs)(
    ...             joblib.delayed(expensive_thing_that_could_be_parallel)()
    ...             for _ in range(3)
    ...         )
    ...
    ...     cheap_span = pewpew.Beam(name="cheap")
    ...     for _ in range(5):
    ...         with cheap_span:
    ...             cheap_serial_thing()
    ...         # simulate some post-processing
    ...         time.sleep(0.01)
    >>>
    >>> def expensive_thing_that_could_be_parallel():
    ...     exp_span = pewpew.Beam(name="expensive_inner")
    ...     with exp_span:
    ...         time.sleep(3)
    >>>
    >>> def cheap_serial_thing():
    ...     time.sleep(random.uniform(0., 0.25))
    >>>
    >>> rethought_function()
    >>> pewpew.draw_trace(annotate=True)
    >>> plt.show()

    Notes
    -----
    * If internal calls to third party libraries that also depend on
      `pewpew` invoke any `Beam` objects, the outermost `trace`
      decorator instance will absorb any collected traces from the
      context object. This holds true for all nested `@trace`-decorated
      functions, internal or not.
    """
    if func is not None:
        val_utils.assert_callable(func)

    def decorated(inner_func):
        try:
            name = inner_func.__name__
        except AttributeError:
            # TODO: can we auto-discover the module?
            name = context_name

        # Internally, the `Decorator` instance returned will encapsulate
        # the callstack in a `TraceContext` object that tracks beams that
        # are created.
        return dec_utils.make_decorator(inner_func, name=name)

    # Code path for `pewpew.pewpew(func=bar, ...)` use case
    if func is not None:
        return decorated(func)

    # Code path for:
    #
    # @pewpew.pewpew
    # def bar(...):
    #     ...
    return decorated



from pathlib import Path

try:
    version_path = Path(__file__).parent / "VERSION"
    version = version_path.read_text().strip()
    del version_path
except FileNotFoundError:
    version = "0.0.0"

__version__ = version
del version
del Path

# submods to include in *
from . import api
from . import beam
from . import tracing
from . import utils

# top level everything
from .api import *

__all__ = [
    # submods
    "api",
    "beam",
    "trace",
    "utils",
]



# -*- coding: utf-8 -*-

"""
Internal logging, primarily used in debugging
"""

import os
import logging

__all__ = [
    "get_logger",
]


def get_logger(name, level=None):
    """Get a PewPew logger"""
    if level is None:
        level = os.environ.get("PEWPEW_LOG_LEVEL", "INFO")

    if not isinstance(level, str):
        raise TypeError("Expected a string for `level`")
    level = logging.getLevelName(level)  # str -> logging level

    log_fmt = "%(asctime)s [%(name)s] [%(levelname)s] %(message)s"
    console_handler = logging.StreamHandler()
    console_handler.setLevel(level)
    console_handler.setFormatter(logging.Formatter(log_fmt))
    logging.basicConfig(
        level=level,
        format=log_fmt,
        datefmt="%m-%d %H:%M:%S",
        handlers=[console_handler],
    )

    return logging.getLogger(name)



# -*- coding: utf-8 -*-

import sys

from pewpew import _context as ctx
from pewpew._logging import get_logger
from pewpew.utils import iterables as iter_utils

logger = get_logger(__name__)


def init_matplotlib(backend=None):
    """Initialize the Matplotlib backend

    In some Mac distros, python is not installed as a framework.
    This means that using the TkAgg backend is the best solution
    (so it doesn't try to use the mac OS backend by default).

    Parameters
    ----------
    backend : str, optional
        The backend to use by default.
    """
    import matplotlib

    existing_backend = matplotlib.get_backend()
    if backend is not None:
        matplotlib.use(backend)
        logger.debug(
            f"Currently using '{existing_backend}' matplotlib backend, "
            f"switching to '{backend}' backend\n"
        )
    else:
        logger.debug(f"Using '{existing_backend}' MPL backend\n")


def draw_trace(
    name=None,
    idx=None,
    title=None,
    **kwargs,
):
    """Draw a `@pewpew.trace`d function

    If `name` or `idx` is not provided, will draw the most recent trace.

    Parameters
    ----------
    name : str, optional
        The name of the trace to draw

    idx : int, optional
        The index of the trace to draw

    title : str, optional
        The optional title of the timeline.
    """
    beams = ctx.ContextStore.get_trace(name=name, idx=idx)
    if not beams:
        raise RuntimeError("No trace history!")
    return _draw_beams(beams, title=title, **kwargs)


def _draw_beams(
    traces,
    color_map="plasma",
    dims=(17.0, None),
    dpi=None,
    alpha=0.66,
    padding=0.001,
    title=None,
    backend=None,
    annotate=False,
    save_to=None,
    save_fmt="png",
    annotate_fontsize="xx-small",
):
    """Plot a number of `Beam` traces on a timeline

    Given a list or tuple of `Beam` traces, plot them on a timeline to display hot
    spots in your code. These can suggest potential areas for parallelism, where
    possible.

    Parameters
    ----------
    color_map : str, optional
        The matplotlib cmap.

    dims : tuple
        A tuple of dimensions.

    dpi : int, optional
        The DPI quality of the image. Default is None.

    alpha : float, optional
        The opacity of the displayed bars. Lower `alpha` means
        more transparent.

    padding : float, optional
        The padding for annotations.

    title : str, optional
        The optional title of the timeline.

    backend : str, optional
        The matplotlib backend to default to.

    annotate : bool, optional
        Whether to annotate the plot. Defaults to False.

    annotate_fontsize : str, optional
        The font size for annotation. Defaults to "xx-small" if not given.

    save_to : str, optional
        A file to save to. If present, the figure will be saved to the location.

    save_fmt : str, optional
        The format of file to save. Defaults to "png"

    Notes
    -----
    * Note that this direct use of this function requires you to collect your
      traces (Ex. 1) which is not the most ergonomic or recommended use.
      Recommended use is through the top-level `@pewpew.trace` decorator
      decorating the outer function, letting the `TraceContext` collect all
      traces, and then calling `pewpew.draw_trace`.

    Examples
    --------
    Example 1, direct use while collecting traces:

    >>> import pewpew
    >>> from pewpew import _plotting as plt_utils
    >>> import time
    >>> import matplotlib.pyplot as plt
    >>>
    >>> outer = pewpew.Beam("outer")
    >>> traces = [outer]
    >>> with outer:
    ...     for _ in range(5):
    ...         inner = pewpew.Beam("inner")
    ...         with inner:
    ...             time.sleep(3)
    ...         traces.append(inner)
    >>> time.sleep(0.5)  # simulate some other process
    >>> with outer:
    ...     time.sleep(2)
    >>> plt_utils.draw_beams(traces, title="example 1", annotate=True)
    >>> plt.show()
    """
    iter_utils.assert_sized_iterable(traces, arg_name="traces", gt_size=0)
    iter_utils.assert_sized_iterable(dims, arg_name="dims", eq_size=2)

    # some os are particular about the backend
    init_matplotlib(backend=backend)

    import matplotlib as mpl
    import matplotlib.pyplot as plt

    n_traces = len(traces)
    cmap = mpl.cm.get_cmap(color_map)
    fig, axes = plt.subplots(n_traces, sharex=True, gridspec_kw={"hspace": 0}, dpi=dpi)

    # if only one trace, make it indexable
    if not hasattr(axes, "__iter__"):
        axes = [axes]

    # set overall title, optionally
    if title:
        fig.suptitle(title)

    # set dims
    if dims[1] is None:
        dims = list(dims[:1]) + [n_traces + 0.5]
    fig.set_size_inches(*dims)

    # Order traces by when their logical blocks were first entered (min start)
    traces = sorted(traces, key=lambda s: s._first_start_time())

    # Get min start time to scale all times relative to zero
    min_start = traces[0]._first_start_time()
    max_stop = max(t._last_end_time() for t in traces)
    max_stop_scaled = (max_stop - min_start) + 0.01
    plt.xlim(-0.01, max_stop_scaled)

    for i, trace in enumerate(traces):
        ax = axes[i]
        ax.set_ylabel(trace.name)
        ax.set_ylim(0, 1)
        ax.set_yticks([])
        ax.set_xlabel("time (s)")
        ax.set_xticklabels([])
        ax.grid(which="both", axis="x", color="k", linestyle=":")

        series = []
        for j, (start, stop) in enumerate(trace._zipped()):
            elapsed = stop - start

            # "For each tuple (xmin, xwidth) a rectangle is drawn from xmin to xmin + xwidth"...
            start_scaled = start - min_start
            series.append((start_scaled, elapsed))

            if annotate:
                annotation = "\n".join(
                    [
                        f"iter: {j}",
                        f"time: {elapsed:,.3f}",
                    ]
                )
                ax.text(
                    start_scaled + padding,
                    0.55 - (0.25 * (j % 2)),
                    annotation,
                    horizontalalignment="left",
                    verticalalignment="center",
                    fontsize=annotate_fontsize,
                )

        ax.broken_barh(
            series, (0, 1), color=cmap(i / n_traces), linewidth=1, alpha=alpha
        )

    if save_to:
        plt.savefig(save_to, format=save_fmt)



# -*- coding: utf-8 -*-

"""
Iterable utils
"""

from pewpew.utils import validation as val_utils

__all__ = [
    "assert_sized_iterable",
    "find_first",
]


def assert_sized_iterable(
    x, *, arg_name=None, lt_size=None, eq_size=None, gt_size=None
):
    """Assert an element is an iterable and of a certain size

    This method is used for arg validation in functions. Rather than
    explicitly include type checking and length checks in all functions,
    this asserts all in one place.

    Parameters
    ----------
    x : iterable or object
        As object to assert is an iterable

    arg_name : str, optional
        The name of the arg

    lt_size : int, optional
        A length to assert an iterable is LESS THAN

    eq_size : int, optional
        A length to assert an iterable exactly EQUALS

    gt_size : int, optional
        A length to assert an iterable is GREATER THAN
    """
    arg_name = val_utils.coalesce(arg_name, "arg")

    if not isinstance(x, (list, tuple)):
        raise TypeError(
            f"Expected list or tuple for `{arg_name}` but got type={type(x)}"
        )

    n_elem = len(x)
    if lt_size is not None and n_elem >= lt_size:
        raise ValueError(
            f"Expected {arg_name} to have fewer than {lt_size} items, but found {n_elem}"
        )

    if eq_size is not None and n_elem != eq_size:
        raise ValueError(
            f"Expected {arg_name} to have exactly {eq_size} items, but found {n_elem}"
        )

    if gt_size is not None and n_elem <= gt_size:
        raise ValueError(
            f"Expected {arg_name} to have greater than {gt_size} items, but found {n_elem}"
        )


def find_first(it, true_fn):
    """Find and return first element where `true_fn` is True

    Finds the first element in `it` for which `true_fn` is True, returning
    the value. Note subsequent elements for which `true_fn` evaluates True
    will not be returned. Elements are evaluated left to right.

    Parameters
    ----------
    it : iterable
        The iterable to evaluate

    true_fn : func
        A function that accepts an element of `it` and evaluates to a
        truthy value.

    Notes
    -----
    * If you pass a generator for `it`, it will be exhausted
    """
    for e in it:
        if true_fn(e):
            return e
    return None



# -*- coding: utf-8 -*-

"""
Validation and type safety
"""

__all__ = [
    "assert_callable",
    "coalesce",
]


def assert_callable(func):
    """Assert an object is a callable"""
    if not callable(func):
        raise TypeError(f"{func} is not a callable object")


def coalesce(x, default):
    """Coalesce a `None` value

    Parameters
    ----------
    x : int, str, object, None or iterable
        An object to coalesce to a non-None value, if None.

    default : int, str, object or iterable
        The non-None value
    """
    if x is None:
        return default
    return x



# -*- coding: utf-8 -*-

"""
Hoist top level imports
"""

from .beam import Beam  # noqa
from .tracing import trace  # noqa
from ._context import clear_trace_history  # noqa
from ._plotting import draw_trace  # noqa

__all__ = [s for s in dir()]



# -*- coding: utf-8 -*-

from pathlib import Path
from setuptools import setup
from setuptools import find_packages

# Get pewpew version
try:
    version_path = Path(__file__).parent / "pewpew" / "VERSION"
    pewpew_version = version_path.read_text().strip()
except FileNotFoundError:
    pewpew_version = "0.0.0"

with open("requirements.txt", "r") as req:
    REQUIREMENTS = [line for line in req.read().split() if line]

print("Requirements:\n" + "\n".join(REQUIREMENTS))
with open("README.md", encoding="utf-8") as readme:
    LONG_DESCRIPTION = readme.read()


def do_setup():
    setup(
        name="pewpew",
        version=pewpew_version,
        description="Cool pythonic utilities for safer, more efficient development",
        maintainer="Taylor G Smith",
        maintainer_email="taylor.smith@alkaline-ml.com",
        license="MIT",
        packages=find_packages(),
        include_package_data=True,
        install_requires=REQUIREMENTS,
        python_requires=">=3.8, <4",
    )


if __name__ == "__main__":
    do_setup()



# -*- coding: utf-8 -*-

import pewpew
import pytest
import time


def test_span():
    span = pewpew.Beam("big_loop")
    with span:
        time.sleep(0.01)

    timings = span.timings
    assert len(timings) == 1, timings


def test_span_concurrent_access_err():
    span = pewpew.Beam("big_loop")
    with span, pytest.raises(RuntimeError):
        _ = span.timings



# -*- coding: utf-8 -*-

import pytest
from pewpew.utils import iterables as iter_utils


@pytest.mark.parametrize(
    "x,arg_name,lt_size,eq_size,gt_size,exp_err,exp_err_msg",
    [
        pytest.param("not an iterable", "x", None, None, None, TypeError, "type="),
        pytest.param(
            [1, 2, 3], None, 2, None, None, ValueError, "arg to have fewer than 2"
        ),
        pytest.param(
            [1, 2, 3], None, None, 4, None, ValueError, "arg to have exactly 4"
        ),
        pytest.param(
            [1, 2, 3], "x", None, None, 3, ValueError, "x to have greater than 3"
        ),
        pytest.param([1, 2, 3], None, None, None, None, None, None),
        pytest.param([1, 2, 3], None, 5, 3, 2, None, None),
    ],
)
def test_assert_sized_iterable(
    x, arg_name, lt_size, eq_size, gt_size, exp_err, exp_err_msg
):
    if exp_err:
        with pytest.raises(exp_err) as e:
            iter_utils.assert_sized_iterable(
                x, arg_name=arg_name, lt_size=lt_size, gt_size=gt_size, eq_size=eq_size
            )
        assert exp_err_msg in str(e)
    else:
        iter_utils.assert_sized_iterable(
            x, arg_name=arg_name, lt_size=lt_size, gt_size=gt_size, eq_size=eq_size
        )


@pytest.mark.parametrize(
    "it,fn,exp",
    [
        pytest.param([], lambda e: e % 2 == 0, None),
        pytest.param([1, 3, 5], lambda e: e % 2 == 0, None),
        pytest.param([2, 4, 6], lambda e: e % 2 == 0, 2),
    ],
)
def test_find_first(it, fn, exp):
    got = iter_utils.find_first(it, fn)
    assert exp == got, got
